@@ -48,7 +48,7 @@ use AnyEvent::Handle ();
 
 use base Exporter::;
 
-our $VERSION = 2.21;
+our $VERSION = 2.24;
 
 our @EXPORT = qw(http_get http_post http_head http_request);
 
@@ -197,6 +197,9 @@ C<$scheme> must be either missing or must be C<http> for HTTP.
 If not specified, then the default proxy is used (see
 C<AnyEvent::HTTP::set_proxy>).
 
+Currently, if your proxy requires authorization, you have to specify an
+appropriate "Proxy-Authorization" header in every request.
+
 =item body => $string
 
 The request body, usually empty. Will be sent as-is (future versions of
@@ -453,7 +456,7 @@ sub _get_slot($$) {
 sub cookie_jar_expire($;$) {
    my ($jar, $session_end) = @_;
 
-   %$jar = () if $jar->{version} != 1;
+   %$jar = () if $jar->{version} != 2;
 
    my $anow = AE::now;
 
@@ -483,20 +486,19 @@ sub cookie_jar_expire($;$) {
 sub cookie_jar_extract($$$$) {
    my ($jar, $scheme, $host, $path) = @_;
 
-   %$jar = () if $jar->{version} != 1;
+   %$jar = () if $jar->{version} != 2;
+
+   $host = AnyEvent::Util::idn_to_ascii $host
+      if $host =~ /[^\x00-\x7f]/;
 
    my @cookies;
 
    while (my ($chost, $paths) = each %$jar) {
       next unless ref $paths;
 
-      if ($chost =~ /^\./) {
-         next unless $chost eq substr $host, -length $chost;
-      } elsif ($chost =~ /\./) {
-         next unless $chost eq $host;
-      } else {
-         next;
-      }
+      # exact match or suffix including . match
+      $chost eq $host or ".$chost" eq substr $host, -1 - length $chost
+         or next;
 
       while (my ($cpath, $cookies) = each %$paths) {
          next unless $cpath eq substr $path, 0, length $cpath;
@@ -528,6 +530,8 @@ sub cookie_jar_extract($$$$) {
 sub cookie_jar_set_cookie($$$$) {
    my ($jar, $set_cookie, $host, $date) = @_;
 
+   %$jar = () if $jar->{version} != 2;
+
    my $anow = int AE::now;
    my $snow; # server-now
 
@@ -582,7 +586,7 @@ sub cookie_jar_set_cookie($$$$) {
       my $cpath = (delete $kv{path}) || "/";
 
       if (exists $kv{domain}) {
-         $cdom = delete $kv{domain};
+         $cdom = $kv{domain};
 
          $cdom =~ s/^\.?/./; # make sure it starts with a "."
 
@@ -591,12 +595,14 @@ sub cookie_jar_set_cookie($$$$) {
          # this is not rfc-like and not netscape-like. go figure.
          my $ndots = $cdom =~ y/.//;
          next if $ndots < ($cdom =~ /\.[^.][^.]\.[^.][^.]$/ ? 3 : 2);
+
+         $cdom = substr $cdom, 1; # remove initial .
       } else {
          $cdom = $host;
       }
 
       # store it
-      $jar->{version} = 1;
+      $jar->{version} = 2;
       $jar->{lc $cdom}{$cpath}{$name} = \%kv;
 
       redo if /\G\s*,/gc;
@@ -772,7 +778,7 @@ sub http_request($$@) {
              : $uscheme eq "https" ? 443
              : return $cb->(undef, { @pseudo, Status => 599, Reason => "Only http and https URL schemes supported" });
 
-   $uauthority =~ /^(?: .*\@ )? ([^\@:]+) (?: : (\d+) )?$/x
+   $uauthority =~ /^(?: .*\@ )? ([^\@]+?) (?: : (\d+) )?$/x
       or return $cb->(undef, { @pseudo, Status => 599, Reason => "Unparsable URL" });
 
    my $uhost = lc $1;
@@ -894,7 +900,7 @@ sub http_request($$@) {
             my $loc = $hdr{location};
 
             if ($loc =~ m%^//%) { # //
-               $loc = "$rscheme:$loc";
+               $loc = "$uscheme:$loc";
 
             } elsif ($loc eq "") {
                $loc = $url;
@@ -903,7 +909,7 @@ sub http_request($$@) {
                $loc =~ s/^\.\/+//;
 
                if ($loc !~ m%^[.?#]%) {
-                  my $prefix = "$rscheme://$uhost:$uport";
+                  my $prefix = "$uscheme://$uauthority";
 
                   unless ($loc =~ s/^\///) {
                      $prefix .= $upath;
@@ -936,11 +942,14 @@ sub http_request($$@) {
             # also, the UA should ask the user for 301 and 307 and POST,
             # industry standard seems to be to simply follow.
             # we go with the industry standard. 308 is defined
-            # by rfc7238
+            # by rfc7538
             if ($status == 301 or $status == 302 or $status == 303) {
-               # HTTP/1.1 is unclear on how to mutate the method
-               $method = "GET" unless $method eq "HEAD";
                $redirect = 1;
+               # HTTP/1.1 is unclear on how to mutate the method
+               unless ($method eq "HEAD") {
+                  $method = "GET";
+                  delete $arg{body};
+               }
             } elsif ($status == 307 or $status == 308) {
                $redirect = 1;
             }
@@ -1027,7 +1036,7 @@ sub http_request($$@) {
          } elsif ($chunked) {
             my $cl = 0;
             my $body = "";
-            my $on_body = $arg{on_body} || sub { $body .= shift; 1 };
+            my $on_body = (!$redirect && $arg{on_body}) || sub { $body .= shift; 1 };
 
             $state{read_chunk} = sub {
                $_[1] =~ /^([0-9a-fA-F]+)/
@@ -1070,7 +1079,7 @@ sub http_request($$@) {
 
             $_[0]->push_read (line => $state{read_chunk});
 
-         } elsif ($arg{on_body}) {
+         } elsif (!$redirect && $arg{on_body}) {
             if (defined $len) {
                $_[0]->on_read (sub {
                   $len -= length $_[0]{rbuf};
@@ -1119,10 +1128,10 @@ sub http_request($$@) {
             %state = ();
             $state{recurse} =
                http_request (
-                  $method   => $url,
+                  $method    => $url,
                   %arg,
-                  recurse   => $recurse - 1,
-                  keepalive => 0,
+                  recurse    => $recurse - 1,
+                  persistent => 0,
                   sub {
                      %state = ();
                      &$cb
@@ -1178,8 +1187,12 @@ sub http_request($$@) {
       if ($proxy && $uscheme eq "https") {
          # oh dear, we have to wrap it into a connect request
 
+         my $auth = exists $hdr{"proxy-authorization"}
+            ? "proxy-authorization: " . (delete $hdr{"proxy-authorization"}) . "\015\012"
+            : "";
+
          # maybe re-use $uauthority with patched port?
-         $state{handle}->push_write ("CONNECT $uhost:$uport HTTP/1.0\015\012\015\012");
+         $state{handle}->push_write ("CONNECT $uhost:$uport HTTP/1.0\015\012$auth\015\012");
          $state{handle}->push_read (line => $qr_nlnl, sub {
             $_[1] =~ /^HTTP\/([0-9\.]+) \s+ ([0-9]{3}) (?: \s+ ([^\015\012]*) )?/ix
                or return _error %state, $cb, { @pseudo, Status => 599, Reason => "Invalid proxy connect response ($_[1])" };
@@ -1192,6 +1205,8 @@ sub http_request($$@) {
             }
          });
       } else {
+         delete $hdr{"proxy-authorization"} unless $proxy;
+
          $handle_actual_request->();
       }
    };
@@ -1207,11 +1222,11 @@ sub http_request($$@) {
          $was_persistent = 1;
 
          $state{handle} = ka_fetch $ka_key;
-         $state{handle}->destroyed
-            and die "AnyEvent::HTTP: unexpectedly got a destructed handle (1), please report.";#d#
+#         $state{handle}->destroyed
+#            and die "AnyEvent::HTTP: unexpectedly got a destructed handle (1), please report.";#d#
          $prepare_handle->();
-         $state{handle}->destroyed
-            and die "AnyEvent::HTTP: unexpectedly got a destructed handle (2), please report.";#d#
+#         $state{handle}->destroyed
+#            and die "AnyEvent::HTTP: unexpectedly got a destructed handle (2), please report.";#d#
          $handle_actual_request->();
 
       } else {
@@ -1279,7 +1294,7 @@ again. If you have a long-running program you can additionally call this
 function from time to time.
 
 A cookie jar is initially an empty hash-reference that is managed by this
-module. It's format is subject to change, but currently it is like this:
+module. Its format is subject to change, but currently it is as follows:
 
 The key C<version> has to contain C<1>, otherwise the hash gets
 emptied. All other keys are hostnames or IP addresses pointing to
@@ -1334,7 +1349,7 @@ C<Mozilla/5.0 (compatible; U; AnyEvent-HTTP/$VERSION; +http://software.schmorp.d
 =item $AnyEvent::HTTP::MAX_PER_HOST
 
 The maximum number of concurrent connections to the same host (identified
-by the hostname). If the limit is exceeded, then the additional requests
+by the hostname). If the limit is exceeded, then additional requests
 are queued until previous connections are closed. Both persistent and
 non-persistent connections are counted in this limit.
 
@@ -1451,8 +1466,6 @@ be retried, and C<1> if it was successful.
       my %hdr;
       my $ofs = 0;
 
-      warn stat $fh;
-      warn -s _;
       if (stat $fh and -s _) {
          $ofs = -s _;
          warn "-s is ", $ofs;
@@ -1490,7 +1503,7 @@ be retried, and C<1> if it was successful.
             my $status = $hdr->{Status};
 
             if (my $time = AnyEvent::HTTP::parse_date $hdr->{"last-modified"}) {
-               utime $fh, $time, $time;
+               utime $time, $time, $fh;
             }
 
             if ($status == 200 || $status == 206 || $status == 416) {
